VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "KVStore"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'=======
'KVStore
'=======
'
'A simple flat Key-Value Store class for persisting simple values like
'program settings to the user's profile.
'
'Hierarchies can be simulated using structured Key names.  Examples:
'
'   Form1_Left
'   Form1_Top
'   Form2_Left
'   Form2_Top
'
'   ABC\DEF\GHI
'
'   Form1 Left
'   Form1 Top
'   Form2 Left
'   Form2 Top
'
'Keys are case-insensitive.
'
'Dependencies:
'
'   o  Shell32.dll 4.71 or later - normally part of Windows since Win95 w/Desktop
'      Update.
'
'KVStore relies on the power of Write#/Input# to make everything locale-blind and
'safe for any language settings.
'
'The Settings Collection holds Variants that are arrays with two elements: the Key
'value and then the Variant Value.  This allows KVStore to persist the Key values
'to disk since a Collection doesn't expose Keys for retrieval.
'
'Using a Default with Item saves space in Settings and in the file, since they won't
'need to be stored.
'
'Persisted data is in the form of Write#/Input# CSV with three fields per record:
'
'   0. Escaped item Key (when hand-editing must be quoted if commas are
'      present).
'   1. Numeric VarType() value of the item.
'   2. Item value.  Strings are in escaped format which allows any Unicode character
'      and all control characters.  When hand-editing String values must be quoted
'      if commas are present. Quotes within values are escaped.
'

Private Enum UnEscStates
    uesNotInEscapeSeq = 0
    uesProcessingEscSeq = 1
    uesProcessingLongEscSeq = 2
    uesProcessingLongEscSeqXEnd = 3
    uesProcessingLongEscSeqUEnd = 5
End Enum

Private LocalAppData As String
Private Settings As Collection

Private mFileName As String 'Value relative to the User's LocalAppData folder.
Private mFullFileName As String

'Not normally used except for debugging:
Public Property Get Count() As Long
    Count = Settings.Count
End Property

Public Property Get FileName() As String
    FileName = mFileName
End Property

Public Property Let FileName(ByVal RHS As String)
    mFileName = RHS
    mFullFileName = LocalAppData & "\" & mFileName
End Property

'Not normally used except for debugging:
Friend Property Get FullFileName() As String
    FullFileName = mFullFileName
End Property

'Not normally used except for debugging:
Friend Property Get ItemNode(ByVal Index As Long) As Variant
    ItemNode = Settings.Item(Index)
End Property

'Default property:
Public Property Get Item(ByVal Key As String, Optional ByVal Default As Variant) As Variant
Attribute Item.VB_UserMemId = 0
    On Error Resume Next
    Item = Settings.Item(Key)(1)
    If Err Then
        Item = Default
    End If
End Property

Public Property Let Item(ByVal Key As String, Optional ByVal Default As Variant, ByVal RHS As Variant)
    On Error Resume Next
    Settings.Remove Key
    On Error GoTo 0
    If Not IsMissing(Default) Then
        If RHS = Default Then RHS = Empty
    End If
    Select Case VarType(RHS)
        Case vbEmpty
            Exit Property 'We're using Empty as a "delete" here.
        Case vbObject, vbError, vbDataObject, vbUserDefinedType
            Err.Raise 380 'Invalid property value.
        Case Else
            If VarType(RHS) And vbArray Then Err.Raise 380 'Invalid property value.
    End Select
    Settings.Add Array(Key, RHS), Key
End Property

Public Function Load() As Boolean
    'Returns True if an existing settings file was found.  If not then you might
    'want to set up "first run" default settings.
    Dim F As Integer
    Dim Key As String
    Dim VType As VbVarType
    Dim Value As Variant
    
    If Len(FileName) = 0 Then Err.Raise &H8004D010, TypeName(Me), "FileName must be set before Load"
    Set Settings = New Collection
    On Error Resume Next
    GetAttr mFullFileName
    If Err.Number = 0 Then
        On Error GoTo 0
        Load = True
        If Settings.Count > 0 Then Set Settings = New Collection
        F = FreeFile(0)
        Open mFullFileName For Input As #F
        Do Until EOF(F)
            Input #F, Key, VType, Value
            Key = UnEscape(Key)
            If VType = vbString Then
                Value = UnEscape(Value)
            Else
                Value = Coerce(Value, VType)
            End If
            Settings.Add Array(Key, Value), Key
        Loop
        Close #F
    End If
End Function

Public Sub Save()
    Dim DirectoryParts() As String
    Dim Path As String
    Dim F As Integer
    Dim Item As Variant
    
    If Len(FileName) = 0 Then Err.Raise &H8004D020, TypeName(Me), "FileName must be set before Save"
    On Error Resume Next
    GetAttr mFullFileName
    If Err Then
        'File not present, so we'll have to create the directory path for the file.
        On Error GoTo 0
        DirectoryParts = Split(mFileName, "\")
        Path = LocalAppData
        If UBound(DirectoryParts) > 0 Then
            For F = 0 To UBound(DirectoryParts) - 1
                Path = Path & "\" & DirectoryParts(F)
                On Error Resume Next
                GetAttr Path
                If Err Then
                    On Error GoTo 0
                    MkDir Path
                Else
                    On Error GoTo 0
                End If
            Next
        End If
    Else
        On Error GoTo 0
    End If
    F = FreeFile(0)
    Open mFullFileName For Output As #F
    For Each Item In Settings
        'By the time we reach here none should be Empty, but test just in case:
        If Not IsEmpty(Item(1)) Then
            Item(0) = Escape(Item(0))
            If VarType(Item(1)) = vbString Then Item(1) = Escape(Item(1))
            Write #F, Item(0), VarType(Item(1)), Item(1)
        End If
    Next
    Close #F
End Sub

Private Function Coerce(ByVal Value As Variant, ByVal VType As VbVarType) As Variant
    Select Case VType
        Case vbNull
            Coerce = Null
        Case vbInteger
            Coerce = CInt(Value)
        Case vbLong
            Coerce = CLng(Value)
        Case vbSingle
            Coerce = CSng(Value)
        Case vbDouble
            Coerce = CDbl(Value)
        Case vbCurrency
            Coerce = CCur(Value)
        Case vbDecimal
            Coerce = CDec(Value)
        Case vbByte
            Coerce = CByte(Value)
        Case Else
            Coerce = Value
    End Select
End Function

Private Function Escape(ByVal PlainText As String) As String
    'Escape a string value.  Escape sequences generated:
    '
    '   ^xhh   7 bit hex
    '   ^uhhhh 16 bit hex ("Unicode")
    '   ^b     vbBack
    '   ^t     vbTab
    '   ^v     vbVerticalTab
    '   ^f     vbFormFeed
    '   ^r     vbCr
    '   ^l     vbLf
    '   ^n     vbNewLine/vbCrLf
    '   ^q     quote (")
    '   ^0     vbNullChar
    '   ^^     ^
    
    Dim Pos As Long
    Dim ChS As String
    Dim ChW As Integer
    Dim EscCh As String
    Dim EscChLen As Long
    Dim EscPos As Long
    
    'Estimate size required as 150% of input length.  Adjust as required.
    Escape = Space$(Len(PlainText) * 15& \ 10&)
    For Pos = 1& To Len(PlainText)
        ChS = Mid$(PlainText, Pos, 1&)
        ChW = AscW(ChS)
        Select Case ChW
            Case 0#
                EscCh = "^0"
            Case 1# To 7#, 14# To 31#, 127#
                EscCh = "^x" & Right$("0" & Hex$(ChW), 2&)
            Case 8#
                EscCh = "^b"
            Case 9#
                EscCh = "^t"
            Case 10#
                EscCh = "^l"
            Case 11#
                EscCh = "^v"
            Case 12#
                EscCh = "^f"
            Case 13#
                If Pos < Len(PlainText) Then
                    If AscW(Mid$(PlainText, Pos + 1, 1&)) = 10# Then
                        EscCh = "^n"
                        Pos = Pos + 1
                    Else
                        EscCh = "^r"
                    End If
                Else
                    EscCh = "^r"
                End If
            Case 32#, 33#, 35# To 93#, 95# To 126#
                EscCh = ChS
            Case 34#
                EscCh = "^q"
            Case 94#
                EscCh = "^^"
            Case Else
                EscCh = "^u" & Right$("000" & Hex$(ChW), 4&)
        End Select
        EscChLen = Len(EscCh)
        'Add room if needed, do it in chunks.
        If Len(Escape) < EscPos + EscChLen Then Escape = Escape & Space$(30&)
        Mid$(Escape, EscPos + 1, EscChLen) = EscCh
        EscPos = EscPos + EscChLen
    Next
    Escape = Left$(Escape, EscPos)
End Function

Private Function UnEscape(ByVal EscapedText As String) As String
    'Un-escape a string value.  Escape sequences supported:
    '
    '   ^xhh   7 bit hex
    '   ^uhhhh 16 bit hex ("Unicode")
    '   ^b     vbBack
    '   ^t     vbTab
    '   ^v     vbVerticalTab
    '   ^f     vbFormFeed
    '   ^r     vbCr
    '   ^l     vbLf
    '   ^n     vbNewLine/vbCrLf
    '   ^q     quote (")
    '   ^0     vbNullChar
    '   ^^     ^
    '
    'Letters indicating escape sequences and hex digits may be upper or lowercase.
    '
    'If an invalid escape sequence is found we raise error 5.
    
    Dim Pos As Long
    Dim ChS As String
    Dim ChW As Integer
    Dim UpChW As Integer
    Dim UnEscState As UnEscStates
    Dim UnPos As Long
    Dim Unicode As Boolean
    Dim Accum As Long
    
    'Estimate output length = input length. Trim back before exit.
    UnEscape = Space$(Len(EscapedText))
    UnPos = 1&
    For Pos = 1& To Len(EscapedText)
        ChS = Mid$(EscapedText, Pos, 1&)
        ChW = AscW(ChS)
        UpChW = AscW(UCase$(ChS))
        Select Case UnEscState
            Case uesNotInEscapeSeq 'Not processing an escape sequence, check for new one.
                If ChW = 94# Then ' "^"
                    UnEscState = uesProcessingEscSeq
                End If
            Case uesProcessingEscSeq 'We're in an escape sequence, check which one.
                'Assume this finishes the sequence until we check further.
                UnEscState = uesNotInEscapeSeq
                Select Case UpChW
                    Case 88# ' "X"
                        UnEscState = uesProcessingLongEscSeq
                    Case 85# ' "U"
                        Unicode = True
                        UnEscState = uesProcessingLongEscSeq
                    Case 66# ' "B"
                        ChS = vbBack
                    Case 84# ' "T"
                        ChS = vbTab
                    Case 86# ' "V"
                        ChS = vbVerticalTab
                    Case 70# ' "F"
                        ChS = vbFormFeed
                    Case 82# ' "R"
                        ChS = vbCr
                    Case 76# ' "L"
                        ChS = vbLf
                    Case 78# ' "N"
                        Mid$(UnEscape, UnPos, 1&) = vbCr
                        UnPos = UnPos + 1 'Add extra here, 2 chars inserted.
                        ChS = vbLf
                    Case 81# ' "Q"
                        ChS = """"
                    Case 48# ' "0"
                        ChS = vbNullChar
                    Case 94# ' "^", i.e. ^^
                        'ChS = ChS
                    Case Else
                        'Undefined escape sequence type char.
                        Err.Raise 5 'Invalid procedure call or argument.
                End Select
            Case Else
                'Working on a ^x or ^u sequence.  UpChW is a hex digit.
                'If ^x we require 2 hex digits, if ^u we require 4.
                
                If (48# <= UpChW And UpChW <= 57#) Then
                    '0 through 9.
                    Accum = 16& * Accum + (CLng(UpChW) And &HF&)
                ElseIf (65# <= UpChW And UpChW <= 70#) Then
                    'A through F.
                    Accum = 16& * Accum + (CLng(UpChW) - 55&) 'Subtracting Asc("A") + 10 here.
                Else
                    'Invalid hex digit.
                    Err.Raise 5
                End If
                'Advance the UnEscState state for this hex digit.
                UnEscState = UnEscState + 1&
                If Unicode Then
                    If UnEscState > uesProcessingLongEscSeqUEnd Then
                        ChS = ChrW$(Accum)
                        Accum = 0&
                        UnEscState = uesNotInEscapeSeq
                        Unicode = False
                    End If
                Else
                    '^x## only valid through ^x7F to avoid ANSI codepage woes.
                    If Accum > 127 Then Err.Raise 5
                    
                    If UnEscState > uesProcessingLongEscSeqXEnd Then
                        ChS = ChrW$(Accum)
                        Accum = 0&
                        UnEscState = uesNotInEscapeSeq
                    End If
                End If
        End Select
        If UnEscState = uesNotInEscapeSeq Then
            Mid$(UnEscape, UnPos, 1&) = ChS
            UnPos = UnPos + 1&
        End If
    Next
    If UnEscState <> uesNotInEscapeSeq Then Err.Raise 5 'Unfinished escape sequence at end.
    UnEscape = Left$(UnEscape, UnPos - 1&)
End Function

Private Sub Class_Initialize()
    Const ssfLOCALAPPDATA = &H1C
    
    With CreateObject("Shell.Application").NameSpace(ssfLOCALAPPDATA).Self
        LocalAppData = .Path
    End With
    Set Settings = New Collection
End Sub
